# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE
from __future__ import annotations

import cupy as cp
import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def test_2651_parameter_union():
    layout = ak.contents.IndexedArray(
        ak.index.Index64([0, 1, 2]),
        ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.uint32)),
        parameters={"foo": {"bar": "baz"}},
    )

    cuda_layout = ak.to_backend(layout, "cuda", highlevel=False)

    result = cuda_layout.project()
    assert result.is_equal_to(
        ak.contents.NumpyArray(
            cp.array([1, 2, 3], dtype=cp.uint32), parameters={"foo": {"bar": "baz"}}
        )
    )


def test_1928_replace_simplify_method_with_classmethod_constructor_indexed_of_union():
    unionarray = ak.from_iter(
        [0.0, 1.1, "zero", 2.2, "one", "two", "three", 3.3, 4.4, 5.5, "four"],
        highlevel=False,
    )

    cuda_unionarray = ak.to_backend(unionarray, "cuda", highlevel=False)

    indexedarray = ak.contents.IndexedArray.simplified(
        ak.index.Index64(cp.array([4, 3, 3, 8, 7, 6], cp.int64)),
        cuda_unionarray,
    )
    assert indexedarray.to_list() == ["one", 2.2, 2.2, 4.4, 3.3, "three"]


def test_1928_replace_simplify_method_with_classmethod_constructor_indexedoption_of_union():
    unionarray = ak.from_iter(
        [0.0, 1.1, "zero", 2.2, "one", "two", "three", 3.3, 4.4, 5.5, "four"],
        highlevel=False,
    )

    cuda_unionarray = ak.to_backend(unionarray, "cuda", highlevel=False)

    indexedoptionarray = ak.contents.IndexedOptionArray.simplified(
        ak.index.Index64(cp.array([-1, 4, 3, -1, 3, 8, 7, 6, -1], cp.int64)),
        cuda_unionarray,
    )
    assert indexedoptionarray.to_list() == [
        None,
        "one",
        2.2,
        None,
        2.2,
        4.4,
        3.3,
        "three",
        None,
    ]


def test_1928_replace_simplify_method_with_classmethod_constructor_indexedoption_of_union_of_option_1():
    with pytest.raises(
        TypeError, match=r" must either be comprised of entirely optional contents"
    ):
        ak.contents.UnionArray(
            ak.index.Index8(cp.array([0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1], dtype=cp.int8)),
            ak.index.Index64(
                cp.array([0, 1, 0, 2, 1, 2, 3, 3, 4, 5, 4], dtype=cp.int64)
            ),
            [
                ak.from_iter([0.0, 1.1, 2.2, 3.3, None, 5.5], highlevel=False),
                ak.from_iter(["zero", "one", "two", "three", "four"], highlevel=False),
            ],
        )

    unionarray = ak.contents.UnionArray.simplified(
        ak.index.Index8(np.array([0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1], dtype=np.int8)),
        ak.index.Index64(np.array([0, 1, 0, 2, 1, 2, 3, 3, 4, 5, 4], dtype=np.int64)),
        [
            ak.from_iter([0.0, 1.1, 2.2, 3.3, None, 5.5], highlevel=False),
            ak.from_iter(["zero", "one", "two", "three", "four"], highlevel=False),
        ],
    )

    cuda_unionarray = ak.to_backend(unionarray, "cuda", highlevel=False)

    indexedoptionarray = ak.contents.IndexedOptionArray.simplified(
        ak.index.Index64(cp.array([-1, 4, 3, -1, 3, 8, 7, 6, -1], cp.int64)),
        cuda_unionarray,
    )
    assert indexedoptionarray.to_list() == [
        None,
        "one",
        2.2,
        None,
        2.2,
        None,
        3.3,
        "three",
        None,
    ]


def test_1928_replace_simplify_method_with_classmethod_constructor_indexedoption_of_union_of_option_2():
    with pytest.raises(
        TypeError, match=r"must either be comprised of entirely optional contents"
    ):
        ak.contents.UnionArray(
            ak.index.Index8(cp.array([0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1], dtype=cp.int8)),
            ak.index.Index64(
                cp.array([0, 1, 0, 2, 1, 2, 3, 3, 4, 5, 4], dtype=cp.int64)
            ),
            [
                ak.from_iter([0.0, 1.1, 2.2, 3.3, 4.4, 5.5], highlevel=False),
                ak.from_iter(["zero", None, "two", "three", "four"], highlevel=False),
            ],
        )

    unionarray = ak.contents.UnionArray.simplified(
        ak.index.Index8(np.array([0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1], dtype=np.int8)),
        ak.index.Index64(np.array([0, 1, 0, 2, 1, 2, 3, 3, 4, 5, 4], dtype=np.int64)),
        [
            ak.from_iter([0.0, 1.1, 2.2, 3.3, 4.4, 5.5], highlevel=False),
            ak.from_iter(["zero", None, "two", "three", "four"], highlevel=False),
        ],
    )

    cuda_unionarray = ak.to_backend(unionarray, "cuda", highlevel=False)

    indexedoptionarray = ak.contents.IndexedOptionArray.simplified(
        ak.index.Index64(cp.array([-1, 4, 3, -1, 3, 8, 7, 6, -1], cp.int64)),
        cuda_unionarray,
    )
    assert indexedoptionarray.to_list() == [
        None,
        None,
        2.2,
        None,
        2.2,
        4.4,
        3.3,
        "three",
        None,
    ]


def test_1928_replace_simplify_method_with_classmethod_constructor_indexedoption_of_union_of_option_1_2():
    unionarray = ak.contents.UnionArray(
        ak.index.Index8(np.array([0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1], dtype=np.int8)),
        ak.index.Index64(np.array([0, 1, 0, 2, 1, 2, 3, 3, 4, 5, 4], dtype=np.int64)),
        [
            ak.from_iter([0.0, 1.1, 2.2, 3.3, None, 5.5], highlevel=False),
            ak.from_iter(["zero", None, "two", "three", "four"], highlevel=False),
        ],
    )

    cuda_unionarray = ak.to_backend(unionarray, "cuda", highlevel=False)

    indexedoptionarray = ak.contents.IndexedOptionArray.simplified(
        ak.index.Index64(cp.array([-1, 4, 3, -1, 3, 8, 7, 6, -1], cp.int64)),
        cuda_unionarray,
    )
    assert indexedoptionarray.to_list() == [
        None,
        None,
        2.2,
        None,
        2.2,
        None,
        3.3,
        "three",
        None,
    ]


def test_0028_add_dressed_types_highlevel():
    a = ak.highlevel.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], check_valid=True
    )
    cuda_a = ak.to_backend(a, "cuda")

    assert (
        repr(cuda_a)
        == "<Array [[1.1, 2.2, 3.3], [], ..., [7.7, 8.8, 9.9]] type='5 * var * float64'>"
    )
    assert str(cuda_a) == "[[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]"

    b = ak.highlevel.Array(np.arange(100, dtype=np.int32), check_valid=True)
    cuda_b = ak.to_backend(b, "cuda")

    assert (
        repr(cuda_b)
        == "<Array [0, 1, 2, 3, 4, 5, 6, ..., 94, 95, 96, 97, 98, 99] type='100 * int32'>"
    )
    assert (
        str(cuda_b)
        == "[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, ..., 90, 91, 92, 93, 94, 95, 96, 97, 98, 99]"
    )

    c = ak.highlevel.Array(
        '[{"one": 3.14, "two": [1.1, 2.2]}, {"one": 99.9, "two": [-3.1415926]}]',
        check_valid=True,
    )
    cuda_c = ak.to_backend(c, "cuda")

    assert (
        repr(cuda_c)
        == "<Array [{one: 3.14, two: [...]}, {...}] type='2 * {one: float64, two: var *...'>"
    )
    assert (
        str(cuda_c) == "[{one: 3.14, two: [1.1, 2.2]}, {one: 99.9, two: [-3.1415926]}]"
    )


def test_0049_distinguish_record_and_recordarray_behaviors():
    class Pointy(ak.highlevel.Record):
        def __str__(self):
            return "<{} {}>".format(self["x"], self["y"])

    behavior = {}
    behavior["__typestr__", "Point"] = "P"
    behavior["Point"] = Pointy
    array = ak.highlevel.Array(
        [
            [{"x": 1, "y": [1.1]}, {"x": 2, "y": [2.0, 0.2]}],
            [],
            [{"x": 3, "y": [3.0, 0.3, 3.3]}],
        ],
        with_name="Point",
        behavior=behavior,
        check_valid=True,
    )
    cuda_array = ak.to_backend(array, "cuda")

    assert str(cuda_array[0][0]) == "<1 [1.1]>"
    assert repr(cuda_array[0]) == "<Array [<1 [1.1]>, <2 [2.0, 0.2]>] type='2 * P'>"
    assert (
        repr(cuda_array)
        == "<Array [[<1 [1.1]>, <2 [2.0, 0.2]>], ..., [{...}]] type='3 * var * P'>"
    )


def test_0074_argsort_and_sort_UnionArray_FIXME():
    content0 = ak.operations.from_iter(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False
    )

    content1 = ak.operations.from_iter(
        ["one", "two", "three", "four", "five"], highlevel=False
    )

    tags = ak.index.Index8([])
    index = ak.index.Index32([])
    array = ak.contents.UnionArray(tags, index, [content0, content1])
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(cuda_array) == []

    assert to_list(ak.operations.sort(cuda_array)) == []
    assert (
        to_list(
            ak.operations.argsort(
                cuda_array,
            )
        )
        == []
    )


def test_0078_argcross_and_cross_axis0():
    one = ak.Array([1.1, 2.2, 3.3])
    two = ak.Array([100, 200, 300, 400, 500])
    three = ak.Array(["a", "b"])
    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")
    cuda_three = ak.to_backend(three, "cuda")

    assert to_list(ak.operations.cartesian([cuda_one], axis=0)) == [
        (1.1,),
        (2.2,),
        (3.3,),
    ]
    assert to_list(ak.operations.cartesian({"x": cuda_one}, axis=0)) == [
        {"x": 1.1},
        {"x": 2.2},
        {"x": 3.3},
    ]

    assert to_list(ak.operations.cartesian([cuda_one, cuda_two], axis=0)) == [
        (1.1, 100),
        (1.1, 200),
        (1.1, 300),
        (1.1, 400),
        (1.1, 500),
        (2.2, 100),
        (2.2, 200),
        (2.2, 300),
        (2.2, 400),
        (2.2, 500),
        (3.3, 100),
        (3.3, 200),
        (3.3, 300),
        (3.3, 400),
        (3.3, 500),
    ]
    assert to_list(ak.operations.cartesian({"x": cuda_one, "y": cuda_two}, axis=0)) == [
        {"x": 1.1, "y": 100},
        {"x": 1.1, "y": 200},
        {"x": 1.1, "y": 300},
        {"x": 1.1, "y": 400},
        {"x": 1.1, "y": 500},
        {"x": 2.2, "y": 100},
        {"x": 2.2, "y": 200},
        {"x": 2.2, "y": 300},
        {"x": 2.2, "y": 400},
        {"x": 2.2, "y": 500},
        {"x": 3.3, "y": 100},
        {"x": 3.3, "y": 200},
        {"x": 3.3, "y": 300},
        {"x": 3.3, "y": 400},
        {"x": 3.3, "y": 500},
    ]
    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], axis=0)
    ) == [
        (1.1, 100, "a"),
        (1.1, 100, "b"),
        (1.1, 200, "a"),
        (1.1, 200, "b"),
        (1.1, 300, "a"),
        (1.1, 300, "b"),
        (1.1, 400, "a"),
        (1.1, 400, "b"),
        (1.1, 500, "a"),
        (1.1, 500, "b"),
        (2.2, 100, "a"),
        (2.2, 100, "b"),
        (2.2, 200, "a"),
        (2.2, 200, "b"),
        (2.2, 300, "a"),
        (2.2, 300, "b"),
        (2.2, 400, "a"),
        (2.2, 400, "b"),
        (2.2, 500, "a"),
        (2.2, 500, "b"),
        (3.3, 100, "a"),
        (3.3, 100, "b"),
        (3.3, 200, "a"),
        (3.3, 200, "b"),
        (3.3, 300, "a"),
        (3.3, 300, "b"),
        (3.3, 400, "a"),
        (3.3, 400, "b"),
        (3.3, 500, "a"),
        (3.3, 500, "b"),
    ]

    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], axis=0, nested=[0])
    ) == [
        [
            (1.1, 100, "a"),
            (1.1, 100, "b"),
            (1.1, 200, "a"),
            (1.1, 200, "b"),
            (1.1, 300, "a"),
        ],
        [
            (1.1, 300, "b"),
            (1.1, 400, "a"),
            (1.1, 400, "b"),
            (1.1, 500, "a"),
            (1.1, 500, "b"),
        ],
        [
            (2.2, 100, "a"),
            (2.2, 100, "b"),
            (2.2, 200, "a"),
            (2.2, 200, "b"),
            (2.2, 300, "a"),
        ],
        [
            (2.2, 300, "b"),
            (2.2, 400, "a"),
            (2.2, 400, "b"),
            (2.2, 500, "a"),
            (2.2, 500, "b"),
        ],
        [
            (3.3, 100, "a"),
            (3.3, 100, "b"),
            (3.3, 200, "a"),
            (3.3, 200, "b"),
            (3.3, 300, "a"),
        ],
        [
            (3.3, 300, "b"),
            (3.3, 400, "a"),
            (3.3, 400, "b"),
            (3.3, 500, "a"),
            (3.3, 500, "b"),
        ],
    ]
    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], axis=0, nested=[1])
    ) == [
        [(1.1, 100, "a"), (1.1, 100, "b")],
        [(1.1, 200, "a"), (1.1, 200, "b")],
        [(1.1, 300, "a"), (1.1, 300, "b")],
        [(1.1, 400, "a"), (1.1, 400, "b")],
        [(1.1, 500, "a"), (1.1, 500, "b")],
        [(2.2, 100, "a"), (2.2, 100, "b")],
        [(2.2, 200, "a"), (2.2, 200, "b")],
        [(2.2, 300, "a"), (2.2, 300, "b")],
        [(2.2, 400, "a"), (2.2, 400, "b")],
        [(2.2, 500, "a"), (2.2, 500, "b")],
        [(3.3, 100, "a"), (3.3, 100, "b")],
        [(3.3, 200, "a"), (3.3, 200, "b")],
        [(3.3, 300, "a"), (3.3, 300, "b")],
        [(3.3, 400, "a"), (3.3, 400, "b")],
        [(3.3, 500, "a"), (3.3, 500, "b")],
    ]
    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], axis=0, nested=[0, 1])
    ) == [
        [
            [(1.1, 100, "a"), (1.1, 100, "b")],
            [(1.1, 200, "a"), (1.1, 200, "b")],
            [(1.1, 300, "a"), (1.1, 300, "b")],
            [(1.1, 400, "a"), (1.1, 400, "b")],
            [(1.1, 500, "a"), (1.1, 500, "b")],
        ],
        [
            [(2.2, 100, "a"), (2.2, 100, "b")],
            [(2.2, 200, "a"), (2.2, 200, "b")],
            [(2.2, 300, "a"), (2.2, 300, "b")],
            [(2.2, 400, "a"), (2.2, 400, "b")],
            [(2.2, 500, "a"), (2.2, 500, "b")],
        ],
        [
            [(3.3, 100, "a"), (3.3, 100, "b")],
            [(3.3, 200, "a"), (3.3, 200, "b")],
            [(3.3, 300, "a"), (3.3, 300, "b")],
            [(3.3, 400, "a"), (3.3, 400, "b")],
            [(3.3, 500, "a"), (3.3, 500, "b")],
        ],
    ]

    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], axis=0, nested=[])
    ) == to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], axis=0, nested=False)
    )
    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], axis=0, nested=[])
    ) == to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], axis=0, nested=None)
    )
    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], axis=0, nested=[0, 1])
    ) == to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], axis=0, nested=True)
    )


def test_0078_argcross_and_cross_axis1():
    one = ak.Array([[0, 1, 2], [], [3, 4]])
    two = ak.Array([[100, 200], [300], [400, 500]])
    three = ak.Array([["a", "b"], ["c", "d"], ["e"]])
    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")
    cuda_three = ak.to_backend(three, "cuda")

    assert to_list(ak.operations.cartesian([cuda_one])) == [
        [(0,), (1,), (2,)],
        [],
        [(3,), (4,)],
    ]
    assert to_list(ak.operations.cartesian({"x": one})) == [
        [{"x": 0}, {"x": 1}, {"x": 2}],
        [],
        [{"x": 3}, {"x": 4}],
    ]

    assert to_list(ak.operations.cartesian([cuda_one, cuda_two])) == [
        [(0, 100), (0, 200), (1, 100), (1, 200), (2, 100), (2, 200)],
        [],
        [(3, 400), (3, 500), (4, 400), (4, 500)],
    ]
    assert to_list(ak.operations.cartesian({"x": one, "y": two})) == [
        [
            {"x": 0, "y": 100},
            {"x": 0, "y": 200},
            {"x": 1, "y": 100},
            {"x": 1, "y": 200},
            {"x": 2, "y": 100},
            {"x": 2, "y": 200},
        ],
        [],
        [
            {"x": 3, "y": 400},
            {"x": 3, "y": 500},
            {"x": 4, "y": 400},
            {"x": 4, "y": 500},
        ],
    ]

    assert to_list(ak.operations.cartesian([cuda_one, cuda_two, cuda_three])) == [
        [
            (0, 100, "a"),
            (0, 100, "b"),
            (0, 200, "a"),
            (0, 200, "b"),
            (1, 100, "a"),
            (1, 100, "b"),
            (1, 200, "a"),
            (1, 200, "b"),
            (2, 100, "a"),
            (2, 100, "b"),
            (2, 200, "a"),
            (2, 200, "b"),
        ],
        [],
        [(3, 400, "e"), (3, 500, "e"), (4, 400, "e"), (4, 500, "e")],
    ]

    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], nested=[0])
    ) == [
        [
            [(0, 100, "a"), (0, 100, "b"), (0, 200, "a"), (0, 200, "b")],
            [(1, 100, "a"), (1, 100, "b"), (1, 200, "a"), (1, 200, "b")],
            [(2, 100, "a"), (2, 100, "b"), (2, 200, "a"), (2, 200, "b")],
        ],
        [],
        [[(3, 400, "e"), (3, 500, "e")], [(4, 400, "e"), (4, 500, "e")]],
    ]
    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], nested=[1])
    ) == [
        [
            [(0, 100, "a"), (0, 100, "b")],
            [(0, 200, "a"), (0, 200, "b")],
            [(1, 100, "a"), (1, 100, "b")],
            [(1, 200, "a"), (1, 200, "b")],
            [(2, 100, "a"), (2, 100, "b")],
            [(2, 200, "a"), (2, 200, "b")],
        ],
        [],
        [[(3, 400, "e")], [(3, 500, "e")], [(4, 400, "e")], [(4, 500, "e")]],
    ]
    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], nested=[0, 1])
    ) == [
        [
            [[(0, 100, "a"), (0, 100, "b")], [(0, 200, "a"), (0, 200, "b")]],
            [[(1, 100, "a"), (1, 100, "b")], [(1, 200, "a"), (1, 200, "b")]],
            [[(2, 100, "a"), (2, 100, "b")], [(2, 200, "a"), (2, 200, "b")]],
        ],
        [],
        [[[(3, 400, "e")], [(3, 500, "e")]], [[(4, 400, "e")], [(4, 500, "e")]]],
    ]

    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], nested=[])
    ) == to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], nested=False)
    )
    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], nested=[])
    ) == to_list(ak.operations.cartesian([cuda_one, cuda_two, cuda_three], nested=None))
    assert to_list(
        ak.operations.cartesian([cuda_one, cuda_two, cuda_three], nested=[0, 1])
    ) == to_list(ak.operations.cartesian([cuda_one, cuda_two, cuda_three], nested=True))


def test_0078_argcross_and_cross_axis2():
    one = ak.Array([[[0, 1, 2], [], [3, 4]], [[0, 1, 2], [], [3, 4]]])
    two = ak.Array([[[100, 200], [300], [400, 500]], [[100, 200], [300], [400, 500]]])
    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")

    assert to_list(ak.operations.cartesian([cuda_one, cuda_two], axis=2)) == [
        [
            [(0, 100), (0, 200), (1, 100), (1, 200), (2, 100), (2, 200)],
            [],
            [(3, 400), (3, 500), (4, 400), (4, 500)],
        ],
        [
            [(0, 100), (0, 200), (1, 100), (1, 200), (2, 100), (2, 200)],
            [],
            [(3, 400), (3, 500), (4, 400), (4, 500)],
        ],
    ]


def test_0078_argcross_and_cross_localindex():
    array = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(ak._do.local_index(cuda_array, 0)) == [0, 1, 2, 3, 4]
    assert to_list(ak._do.local_index(cuda_array, 1)) == [
        [0, 1, 2],
        [],
        [0, 1],
        [0],
        [0, 1, 2, 3],
    ]

    array = ak.operations.from_iter(
        [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], [[5.5]], [[6.6, 7.7, 8.8, 9.9]]],
        highlevel=False,
    )
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(ak._do.local_index(cuda_array, 0)) == [0, 1, 2, 3]
    assert to_list(ak._do.local_index(cuda_array, 1)) == [[0, 1, 2], [], [0], [0]]
    assert to_list(ak._do.local_index(cuda_array, 2)) == [
        [[0, 1, 2], [], [0, 1]],
        [],
        [[0]],
        [[0, 1, 2, 3]],
    ]

    array = ak.operations.from_numpy(
        np.arange(2 * 3 * 5).reshape(2, 3, 5), regulararray=True, highlevel=False
    )
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(ak._do.local_index(cuda_array, 0)) == [0, 1]
    assert to_list(ak._do.local_index(cuda_array, 1)) == [[0, 1, 2], [0, 1, 2]]
    assert to_list(ak._do.local_index(cuda_array, 2)) == [
        [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]],
        [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]],
    ]


def test_0078_argcross_and_cross_argcartesian():
    one = ak.Array([[0.0, 1.1, 2.2], [], [3.3, 4.4]])
    two = ak.Array([[100, 200], [300], [400, 500]])
    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")

    assert to_list(ak.operations.argcartesian([cuda_one, cuda_two])) == [
        [(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1)],
        [],
        [(0, 0), (0, 1), (1, 0), (1, 1)],
    ]
    assert to_list(ak.operations.argcartesian({"x": one, "y": two})) == [
        [
            {"x": 0, "y": 0},
            {"x": 0, "y": 1},
            {"x": 1, "y": 0},
            {"x": 1, "y": 1},
            {"x": 2, "y": 0},
            {"x": 2, "y": 1},
        ],
        [],
        [{"x": 0, "y": 0}, {"x": 0, "y": 1}, {"x": 1, "y": 0}, {"x": 1, "y": 1}],
    ]


def test_0078_argcross_and_cross_argcartesian_negative_axis():
    one = ak.Array([[["a", "b"], []], [], [["c"]]])
    two = ak.Array([[[1.1], []], [], [[2.2, 3.3]]])
    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")

    assert ak.cartesian([cuda_one, cuda_two], axis=-1).to_list() == [
        [[("a", 1.1), ("b", 1.1)], []],
        [],
        [[("c", 2.2), ("c", 3.3)]],
    ]

    assert ak.argcartesian([cuda_one, cuda_two], axis=-1).to_list() == [
        [[(0, 0), (1, 0)], []],
        [],
        [[(0, 0), (0, 1)]],
    ]


def test_0093_simplify_uniontypes_and_optiontypes_concatenate():
    one = ak.highlevel.Array([1.1, 2.2, 3.3, 4.4, 5.5], check_valid=True).layout
    two = ak.highlevel.Array([[], [1], [2, 2], [3, 3, 3]], check_valid=True).layout
    three = ak.highlevel.Array(
        [True, False, False, True, True], check_valid=True
    ).layout
    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")
    cuda_three = ak.to_backend(three, "cuda")

    assert to_list(ak.operations.concatenate([cuda_one, cuda_two, cuda_three])) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        [],
        [1],
        [2, 2],
        [3, 3, 3],
        1.0,
        0.0,
        0.0,
        1.0,
        1.0,
    ]
    assert isinstance(
        ak.operations.concatenate([cuda_one, cuda_two, cuda_three], highlevel=False),
        ak.contents.unionarray.UnionArray,
    )
    assert (
        len(
            ak.operations.concatenate(
                [cuda_one, cuda_two, cuda_three], highlevel=False
            ).contents
        )
        == 2
    )


def test_0093_simplify_uniontypes_and_optiontypes_indexedarray_merge():
    content1 = ak.operations.from_iter(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False
    )
    content2 = ak.operations.from_iter([[1, 2], [], [3, 4]], highlevel=False)
    index1 = ak.index.Index64(np.array([2, 0, -1, 0, 1, 2], dtype=np.int64))

    cuda_content2 = ak.to_backend(content2, "cuda", highlevel=False)
    indexedarray1 = ak.contents.IndexedOptionArray(index1, content1)
    cuda_indexedarray1 = ak.to_backend(indexedarray1, "cuda", highlevel=False)

    assert to_list(cuda_indexedarray1) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
    ]

    assert to_list(cuda_indexedarray1._mergemany([cuda_content2])) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [1.0, 2.0],
        [],
        [3.0, 4.0],
    ]
    assert to_list(cuda_content2._mergemany([cuda_indexedarray1])) == [
        [1.0, 2.0],
        [],
        [3.0, 4.0],
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
    ]
    assert to_list(cuda_indexedarray1._mergemany([cuda_indexedarray1])) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
    ]

    assert (
        cuda_indexedarray1.to_typetracer()
        ._mergemany([cuda_content2.to_typetracer()])
        .form
        == cuda_indexedarray1._mergemany([cuda_content2]).form
    )
    assert (
        cuda_content2.to_typetracer()
        ._mergemany([cuda_indexedarray1.to_typetracer()])
        .form
        == cuda_content2._mergemany([cuda_indexedarray1]).form
    )
    assert (
        cuda_indexedarray1.to_typetracer()
        ._mergemany([cuda_indexedarray1.to_typetracer()])
        .form
        == cuda_indexedarray1._mergemany([cuda_indexedarray1]).form
    )


def test_0093_simplify_uniontypes_and_optiontypes_indexedarray_simplify():
    array = ak.operations.from_iter(
        ["one", "two", None, "three", None, None, "four", "five"], highlevel=False
    )
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    index2 = ak.index.Index64(np.array([2, 2, 1, 6, 5], dtype=np.int64))

    array2 = ak.contents.IndexedArray.simplified(index2, array)
    cuda_array2 = ak.to_backend(array2, "cuda", highlevel=False)

    assert cp.asarray(cuda_array.index.data).tolist() == [0, 1, -1, 2, -1, -1, 3, 4]
    assert (
        to_list(cuda_array2)
        == to_list(cuda_array2)
        == [None, None, "two", "four", None]
    )

    assert cuda_array2.to_typetracer().form == cuda_array2.form


def test_0093_simplify_uniontypes_and_optiontypes_where():
    condition = ak.highlevel.Array(
        [True, False, True, False, True],
        check_valid=True,
    )
    one = ak.highlevel.Array([1.1, 2.2, 3.3, 4.4, 5.5], check_valid=True)
    two = ak.highlevel.Array([False, False, False, True, True], check_valid=True)
    three = ak.highlevel.Array(
        [[], [1], [2, 2], [3, 3, 3], [4, 4, 4, 4]], check_valid=True
    )

    cuda_condition = ak.to_backend(condition, "cuda")
    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")
    cuda_three = ak.to_backend(three, "cuda")

    assert to_list(ak.operations.where(cuda_condition, cuda_one, cuda_two)) == [
        1.1,
        0.0,
        3.3,
        1.0,
        5.5,
    ]
    assert to_list(ak.operations.where(cuda_condition, cuda_one, cuda_three)) == [
        1.1,
        [1],
        3.3,
        [3, 3, 3],
        5.5,
    ]


def test_0093_simplify_uniontypes_and_optiontypes_unionarray_simplify_one():
    one = ak.operations.from_iter([5, 4, 3, 2, 1], highlevel=False)
    two = ak.operations.from_iter([[], [1], [2, 2], [3, 3, 3]], highlevel=False)
    three = ak.operations.from_iter([1.1, 2.2, 3.3], highlevel=False)
    tags = ak.index.Index8(
        np.array([0, 0, 1, 2, 1, 0, 2, 1, 1, 0, 2, 0], dtype=np.int8)
    )
    index = ak.index.Index64(
        np.array([0, 1, 0, 0, 1, 2, 1, 2, 3, 3, 2, 4], dtype=np.int64)
    )
    array = ak.contents.UnionArray.simplified(tags, index, [one, two, three])

    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(cuda_array) == [
        5,
        4,
        [],
        1.1,
        [1],
        3,
        2.2,
        [2, 2],
        [3, 3, 3],
        2,
        3.3,
        1,
    ]
    assert len(cuda_array.contents) == 2
    assert cuda_array.to_typetracer().form == cuda_array.form


def test_0093_simplify_uniontypes_and_optiontypes_unionarray_simplify():
    one = ak.operations.from_iter([5, 4, 3, 2, 1], highlevel=False)
    two = ak.operations.from_iter([[], [1], [2, 2], [3, 3, 3]], highlevel=False)
    three = ak.operations.from_iter([1.1, 2.2, 3.3], highlevel=False)

    tags2 = ak.index.Index8(np.array([0, 1, 0, 1, 0, 0, 1], dtype=np.int8))
    index2 = ak.index.Index32(np.array([0, 0, 1, 1, 2, 3, 2], dtype=np.int32))
    inner = ak.contents.UnionArray(tags2, index2, [two, three])
    tags1 = ak.index.Index8(
        np.array([0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 0], dtype=np.int8)
    )
    index1 = ak.index.Index64(
        np.array([0, 1, 0, 1, 2, 2, 3, 4, 5, 3, 6, 4], dtype=np.int64)
    )
    outer = ak.contents.UnionArray.simplified(tags1, index1, [one, inner])
    cuda_outer = ak.to_backend(outer, "cuda", highlevel=False)

    assert to_list(cuda_outer) == [
        5,
        4,
        [],
        1.1,
        [1],
        3,
        2.2,
        [2, 2],
        [3, 3, 3],
        2,
        3.3,
        1,
    ]

    assert isinstance(cuda_outer.content(0), ak.contents.numpyarray.NumpyArray)
    assert isinstance(
        cuda_outer.content(1), ak.contents.listoffsetarray.ListOffsetArray
    )
    assert len(cuda_outer.contents) == 2
    assert cuda_outer.to_typetracer().form == cuda_outer.form

    tags2 = ak.index.Index8(np.array([0, 1, 0, 1, 0, 0, 1], dtype=np.int8))
    index2 = ak.index.Index64(np.array([0, 0, 1, 1, 2, 3, 2], dtype=np.int64))
    inner = ak.contents.UnionArray(tags2, index2, [two, three])
    tags1 = ak.index.Index8(
        np.array([1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1], dtype=np.int8)
    )
    index1 = ak.index.Index32(
        np.array([0, 1, 0, 1, 2, 2, 3, 4, 5, 3, 6, 4], dtype=np.int32)
    )
    outer = ak.contents.UnionArray.simplified(tags1, index1, [inner, one])
    cuda_outer = ak.to_backend(outer, "cuda", highlevel=False)

    assert to_list(cuda_outer) == [
        5,
        4,
        [],
        1.1,
        [1],
        3,
        2.2,
        [2, 2],
        [3, 3, 3],
        2,
        3.3,
        1,
    ]


def test_0093_simplify_uniontypes_and_optiontypes_merge_parameters():
    one = ak.operations.from_iter(
        [[121, 117, 99, 107, 121], [115, 116, 117, 102, 102]], highlevel=False
    )
    two = ak.operations.from_iter(["good", "stuff"], highlevel=False)
    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")

    assert to_list(ak.operations.concatenate([cuda_one, cuda_two])) == [
        [121, 117, 99, 107, 121],
        [115, 116, 117, 102, 102],
        "good",
        "stuff",
    ]
    assert to_list(ak.operations.concatenate([cuda_two, cuda_one])) == [
        "good",
        "stuff",
        [121, 117, 99, 107, 121],
        [115, 116, 117, 102, 102],
    ]

    assert (
        ak.operations.concatenate([cuda_one, cuda_two], highlevel=False)
        .to_typetracer()
        .form
        == ak.operations.concatenate([cuda_one, cuda_two], highlevel=False).form
    )
    assert (
        ak.operations.concatenate([cuda_two, cuda_one], highlevel=False)
        .to_typetracer()
        .form
        == ak.operations.concatenate([cuda_two, cuda_one], highlevel=False).form
    )


def test_0093_simplify_uniontypes_and_optiontypes_unionarray_merge():
    emptyarray = ak.contents.EmptyArray()

    one = ak.operations.from_iter([0.0, 1.1, 2.2, [], [1], [2, 2]], highlevel=False)
    two = ak.operations.from_iter(
        [{"x": 1, "y": 1.1}, 999, 123, {"x": 2, "y": 2.2}], highlevel=False
    )
    three = ak.operations.from_iter(["one", "two", "three"], highlevel=False)

    cuda_emptyarray = ak.to_backend(emptyarray, "cuda", highlevel=False)
    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two])) == [
        0.0,
        1.1,
        2.2,
        [],
        [1],
        [2, 2],
        {"x": 1, "y": 1.1},
        999,
        123,
        {"x": 2, "y": 2.2},
    ]
    assert to_list(cuda_two._mergemany([cuda_one])) == [
        {"x": 1, "y": 1.1},
        999,
        123,
        {"x": 2, "y": 2.2},
        0.0,
        1.1,
        2.2,
        [],
        [1],
        [2, 2],
    ]

    assert to_list(cuda_one._mergemany([cuda_emptyarray])) == [
        0.0,
        1.1,
        2.2,
        [],
        [1],
        [2, 2],
    ]
    assert to_list(cuda_emptyarray._mergemany([cuda_one])) == [
        0.0,
        1.1,
        2.2,
        [],
        [1],
        [2, 2],
    ]

    assert to_list(cuda_one._mergemany([cuda_three])) == [
        0.0,
        1.1,
        2.2,
        [],
        [1],
        [2, 2],
        "one",
        "two",
        "three",
    ]
    assert to_list(cuda_two._mergemany([cuda_three])) == [
        {"x": 1, "y": 1.1},
        999,
        123,
        {"x": 2, "y": 2.2},
        "one",
        "two",
        "three",
    ]
    assert to_list(cuda_three._mergemany([cuda_one])) == [
        "one",
        "two",
        "three",
        0.0,
        1.1,
        2.2,
        [],
        [1],
        [2, 2],
    ]
    assert to_list(cuda_three._mergemany([cuda_two])) == [
        "one",
        "two",
        "three",
        {"x": 1, "y": 1.1},
        999,
        123,
        {"x": 2, "y": 2.2},
    ]

    assert (
        cuda_one.to_typetracer()._mergemany([cuda_two.to_typetracer()]).form
        == cuda_one._mergemany([cuda_two]).form
    )
    assert (
        cuda_two.to_typetracer()._mergemany([cuda_one.to_typetracer()]).form
        == cuda_two._mergemany([cuda_one]).form
    )
    assert (
        cuda_one.to_typetracer()._mergemany([cuda_emptyarray.to_typetracer()]).form
        == cuda_one._mergemany([cuda_emptyarray]).form
    )
    assert (
        cuda_emptyarray.to_typetracer()._mergemany([cuda_one.to_typetracer()]).form
        == cuda_emptyarray._mergemany([cuda_one]).form
    )
    assert (
        cuda_one.to_typetracer()._mergemany([cuda_three.to_typetracer()]).form
        == cuda_one._mergemany([cuda_three]).form
    )
    assert (
        cuda_two.to_typetracer()._mergemany([cuda_three.to_typetracer()]).form
        == cuda_two._mergemany([cuda_three]).form
    )
    assert (
        cuda_three.to_typetracer()._mergemany([cuda_one.to_typetracer()]).form
        == cuda_three._mergemany([cuda_one]).form
    )
    assert (
        cuda_three.to_typetracer()._mergemany([cuda_two.to_typetracer()]).form
        == cuda_three._mergemany([cuda_two]).form
    )


def test_0093_simplify_uniontypes_and_optiontypes_numpyarray_merge():
    emptyarray = ak.contents.EmptyArray()

    np1 = np.arange(2 * 7 * 5).reshape(2, 7, 5)
    np2 = np.arange(3 * 7 * 5).reshape(3, 7, 5)

    cuda_emptyarray = ak.to_backend(emptyarray, "cuda", highlevel=False)
    cuda_np1 = ak.to_backend(np1, "cuda")
    cuda_np2 = ak.to_backend(np2, "cuda")
    cuda_ak1 = ak.contents.NumpyArray(cuda_np1)
    cuda_ak2 = ak.contents.NumpyArray(cuda_np2)

    assert to_list(cuda_ak1._mergemany([cuda_ak2])) == to_list(
        np.concatenate([cuda_np1, cuda_np2])
    )
    assert to_list(
        cuda_ak1[1:, :-1, ::-1]._mergemany([cuda_ak2[1:, :-1, ::-1]])
    ) == to_list(np.concatenate([cuda_np1[1:, :-1, ::-1], cuda_np2[1:, :-1, ::-1]]))
    assert (
        cuda_ak1.to_typetracer()._mergemany([cuda_ak2.to_typetracer()]).form
        == cuda_ak1._mergemany([cuda_ak2]).form
    )
    assert (
        cuda_ak1[1:, :-1, ::-1]
        .to_typetracer()
        ._mergemany([cuda_ak2[1:, :-1, ::-1].to_typetracer()])
        .form
        == cuda_ak1[1:, :-1, ::-1]._mergemany([cuda_ak2[1:, :-1, ::-1]]).form
    )

    for x in [
        np.bool_,
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
        np.float32,
        np.float64,
    ]:
        for y in [
            np.bool_,
            np.int8,
            np.int16,
            np.int32,
            np.int64,
            np.uint8,
            np.uint16,
            np.uint32,
            np.uint64,
            np.float32,
            np.float64,
        ]:
            z = np.concatenate(
                [np.array([1, 2, 3], dtype=x), np.array([4, 5], dtype=y)]
            ).dtype.type
            one = ak.contents.NumpyArray(np.array([1, 2, 3], dtype=x))
            two = ak.contents.NumpyArray(np.array([4, 5], dtype=y))
            cuda_one = ak.to_backend(one, "cuda", highlevel=False)
            cuda_two = ak.to_backend(two, "cuda", highlevel=False)

            cuda_three = cuda_one._mergemany([cuda_two])
            assert ak.to_numpy(cuda_three).dtype == np.dtype(z), (
                f"{x} {y} {z} {ak.to_numpy(cuda_three).dtype.type}"
            )
            assert to_list(cuda_three) == to_list(
                np.concatenate([ak.to_numpy(cuda_one), ak.to_numpy(two)])
            )
            assert to_list(cuda_one._mergemany([cuda_emptyarray])) == to_list(cuda_one)
            assert to_list(cuda_emptyarray._mergemany([cuda_one])) == to_list(cuda_one)

            assert (
                cuda_one.to_typetracer()._mergemany([cuda_two.to_typetracer()]).form
                == cuda_one._mergemany([cuda_two]).form
            )
            assert (
                cuda_one.to_typetracer()
                ._mergemany([cuda_emptyarray.to_typetracer()])
                .form
                == cuda_one._mergemany([cuda_emptyarray]).form
            )
            assert (
                cuda_emptyarray.to_typetracer()
                ._mergemany([cuda_one.to_typetracer()])
                .form
                == cuda_emptyarray._mergemany([cuda_one]).form
            )


def test_0093_simplify_uniontypes_and_optiontypes_regulararray_merge():
    emptyarray = ak.contents.EmptyArray()

    np1 = np.arange(2 * 7 * 5).reshape(2, 7, 5)
    np2 = np.arange(3 * 7 * 5).reshape(3, 7, 5)

    cuda_emptyarray = ak.to_backend(emptyarray, "cuda", highlevel=False)
    cuda_np1 = ak.to_backend(np1, "cuda")
    cuda_np2 = ak.to_backend(np2, "cuda")
    cuda_ak1 = ak.operations.from_iter(cuda_np1, highlevel=False)
    cuda_ak2 = ak.operations.from_iter(cuda_np2, highlevel=False)

    assert to_list(cuda_ak1._mergemany([cuda_ak2])) == to_list(
        np.concatenate([cuda_np1, cuda_np2])
    )
    assert to_list(cuda_ak1._mergemany([cuda_emptyarray])) == to_list(cuda_ak1)
    assert to_list(cuda_emptyarray._mergemany([cuda_ak1])) == to_list(cuda_ak1)

    assert (
        cuda_ak1.to_typetracer()._mergemany([cuda_ak2.to_typetracer()]).form
        == cuda_ak1._mergemany([cuda_ak2]).form
    )
    assert (
        cuda_ak1.to_typetracer()._mergemany([cuda_emptyarray.to_typetracer()]).form
        == cuda_ak1._mergemany([cuda_emptyarray]).form
    )
    assert (
        cuda_emptyarray.to_typetracer()._mergemany([cuda_ak1.to_typetracer()]).form
        == cuda_emptyarray._mergemany([cuda_ak1]).form
    )


def test_0093_simplify_uniontypes_and_optiontypes_mask_as_bool():
    array = ak.operations.from_iter(
        ["one", "two", None, "three", None, None, "four"], highlevel=False
    )
    index2 = ak.index.Index64(np.array([2, 2, 1, 5, 0], dtype=np.int64))
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)
    cuda_index2 = ak.to_backend(index2, "cuda", highlevel=False)

    cuda_array2 = ak.contents.IndexedArray.simplified(cuda_index2, cuda_array)

    assert cp.asarray(
        cuda_array.mask_as_bool(valid_when=False).view(cp.int8)
    ).tolist() == [
        0,
        0,
        1,
        0,
        1,
        1,
        0,
    ]
    assert cp.asarray(
        cuda_array2.mask_as_bool(valid_when=False).view(cp.int8)
    ).tolist() == [
        1,
        1,
        0,
        1,
        0,
    ]


def test_0093_simplify_uniontypes_and_optiontypes_listarray_merge():
    emptyarray = ak.contents.EmptyArray()
    cuda_emptyarray = ak.to_backend(emptyarray, "cuda", highlevel=False)

    content1 = ak.contents.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    content2 = ak.contents.NumpyArray(np.array([1, 2, 3, 4, 5, 6, 7]))

    for (dtype1, Index1, ListArray1), (dtype2, Index2, ListArray2) in [
        (
            (np.int32, ak.index.Index32, ak.contents.ListArray),
            (np.int32, ak.index.Index32, ak.contents.ListArray),
        ),
        (
            (np.int32, ak.index.Index32, ak.contents.ListArray),
            (np.uint32, ak.index.IndexU32, ak.contents.ListArray),
        ),
        (
            (np.int32, ak.index.Index32, ak.contents.ListArray),
            (np.int64, ak.index.Index64, ak.contents.ListArray),
        ),
        (
            (np.uint32, ak.index.IndexU32, ak.contents.ListArray),
            (np.int32, ak.index.Index32, ak.contents.ListArray),
        ),
        (
            (np.uint32, ak.index.IndexU32, ak.contents.ListArray),
            (np.uint32, ak.index.IndexU32, ak.contents.ListArray),
        ),
        (
            (np.uint32, ak.index.IndexU32, ak.contents.ListArray),
            (np.int64, ak.index.Index64, ak.contents.ListArray),
        ),
        (
            (np.int64, ak.index.Index64, ak.contents.ListArray),
            (np.int32, ak.index.Index32, ak.contents.ListArray),
        ),
        (
            (np.int64, ak.index.Index64, ak.contents.ListArray),
            (np.uint32, ak.index.IndexU32, ak.contents.ListArray),
        ),
        (
            (np.int64, ak.index.Index64, ak.contents.ListArray),
            (np.int64, ak.index.Index64, ak.contents.ListArray),
        ),
    ]:
        starts1 = Index1(np.array([0, 3, 3], dtype=dtype1))
        stops1 = Index1(np.array([3, 3, 5], dtype=dtype1))
        starts2 = Index2(np.array([2, 99, 0], dtype=dtype2))
        stops2 = Index2(np.array([6, 99, 3], dtype=dtype2))
        array1 = ListArray1(starts1, stops1, content1)
        array2 = ListArray2(starts2, stops2, content2)

        cuda_array1 = ak.to_backend(array1, "cuda", highlevel=False)
        cuda_array2 = ak.to_backend(array2, "cuda", highlevel=False)

        assert to_list(cuda_array1) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
        assert to_list(cuda_array2) == [[3, 4, 5, 6], [], [1, 2, 3]]

        assert to_list(cuda_array1._mergemany([cuda_array2])) == [
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
            [3, 4, 5, 6],
            [],
            [1, 2, 3],
        ]
        assert to_list(cuda_array2._mergemany([cuda_array1])) == [
            [3, 4, 5, 6],
            [],
            [1, 2, 3],
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
        ]
        assert to_list(cuda_array1._mergemany([cuda_emptyarray])) == to_list(
            cuda_array1
        )
        assert to_list(cuda_emptyarray._mergemany([cuda_array1])) == to_list(
            cuda_array1
        )

        assert (
            cuda_array1.to_typetracer()._mergemany([cuda_array2.to_typetracer()]).form
            == cuda_array1._mergemany([cuda_array2]).form
        )
        assert (
            cuda_array2.to_typetracer()._mergemany([cuda_array1.to_typetracer()]).form
            == cuda_array2._mergemany([cuda_array1]).form
        )
        assert (
            cuda_array1.to_typetracer()
            ._mergemany([cuda_emptyarray.to_typetracer()])
            .form
            == cuda_array1._mergemany([cuda_emptyarray]).form
        )
        assert (
            cuda_emptyarray.to_typetracer()
            ._mergemany([cuda_array1.to_typetracer()])
            .form
            == cuda_emptyarray._mergemany([cuda_array1]).form
        )

    regulararray = ak.contents.RegularArray(content2, 2, zeros_length=0)
    cuda_regulararray = ak.to_backend(regulararray, "cuda", highlevel=False)

    assert to_list(cuda_regulararray) == [[1, 2], [3, 4], [5, 6]]
    assert to_list(cuda_regulararray._mergemany([cuda_emptyarray])) == to_list(
        cuda_regulararray
    )
    assert to_list(cuda_emptyarray._mergemany([cuda_regulararray])) == to_list(
        cuda_regulararray
    )

    for dtype1, Index1, ListArray1 in [
        (np.int32, ak.index.Index32, ak.contents.ListArray),
        (np.uint32, ak.index.IndexU32, ak.contents.ListArray),
        (np.int64, ak.index.Index64, ak.contents.ListArray),
    ]:
        starts1 = Index1(np.array([0, 3, 3], dtype=dtype1))
        stops1 = Index1(np.array([3, 3, 5], dtype=dtype1))
        array1 = ListArray1(starts1, stops1, content1)
        cuda_array1 = ak.to_backend(array1, "cuda", highlevel=False)

        assert to_list(cuda_array1._mergemany([cuda_regulararray])) == [
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
            [1, 2],
            [3, 4],
            [5, 6],
        ]
        assert to_list(cuda_regulararray._mergemany([cuda_array1])) == [
            [1, 2],
            [3, 4],
            [5, 6],
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
        ]


def test_0193_is_none_axis_parameter():
    one = ak.operations.from_iter([1, None, 3], highlevel=False)
    two = ak.operations.from_iter([[], [1], None, [3, 3, 3]], highlevel=False)
    tags = ak.index.Index8(np.array([0, 1, 1, 0, 0, 1, 1], dtype=np.int8))
    index = ak.index.Index64(np.array([0, 0, 1, 1, 2, 2, 3], dtype=np.int64))
    array = ak.Array(ak.contents.UnionArray(tags, index, [one, two]), check_valid=True)
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.to_list(cuda_array) == [1, [], [1], None, 3, None, [3, 3, 3]]

    assert ak.to_list(ak.operations.is_none(cuda_array)) == [
        False,
        False,
        False,
        True,
        False,
        True,
        False,
    ]


def test_0023_regular_array_getitem():
    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.contents.ListOffsetArray(offsets, content)
    regulararray = ak.contents.RegularArray(listoffsetarray, 2, zeros_length=0)

    cuda_regulararray = ak.to_backend(regulararray, "cuda", highlevel=False)

    assert to_list(cuda_regulararray[(0)]) == [[0.0, 1.1, 2.2], []]
    assert to_list(cuda_regulararray[(1)]) == [[3.3, 4.4], [5.5]]
    assert to_list(cuda_regulararray[(2)]) == [[6.6, 7.7, 8.8, 9.9], []]
    assert cuda_regulararray.to_typetracer()[(2)].form == cuda_regulararray[(2)].form
    assert to_list(cuda_regulararray[(slice(1, None, None))]) == [
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]
    assert (
        cuda_regulararray.to_typetracer()[(slice(1, None, None))].form
        == cuda_regulararray[(slice(1, None, None))].form
    )
    assert to_list(cuda_regulararray[(slice(None, -1, None))]) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
    ]
    assert (
        cuda_regulararray.to_typetracer()[(slice(None, -1, None))].form
        == cuda_regulararray[(slice(None, -1, None))].form
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak

# from awkward.operations import to_list


def test_concatenate_as_reported():
    a = ak.Array([[1]])
    a_concat_128 = ak.concatenate([a for i in range(128)], axis=1)
    assert a_concat_128.to_list() == [[1] * 128]

    a_concat_129 = ak.concatenate([a for i in range(129)], axis=1)
    assert a_concat_129.to_list() == [[1] * 129]


def test_concatenate_inner_union_simplify_one():
    a = ak.Array([[99]])
    astr = ak.Array(["a b c d".split()])
    aa = [a for i in range(129)] + [astr]

    cu = ak.concatenate(aa, axis=1)
    assert cu.to_list() == [[99] * 129 + ["a", "b", "c", "d"]]


def test_concatenate_inner_union_simplify():
    a = ak.Array([[99]])
    amulti = ak.Array([[1, 2, "a", "b"]])
    aa = [a for i in range(129)] + [amulti]

    cu = ak.concatenate(aa, axis=1)
    assert cu.to_list() == [[99] * 129 + [1, 2, "a", "b"]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test_ak_zip_no_broadcast_NumpyArray_dict():
    a = ak.Array([1])
    b = ak.Array([2])
    c = ak.zip_no_broadcast({"a": a, "b": b})
    assert ak.to_list(c) == ak.to_list(ak.zip({"a": a, "b": b}))


def test_ak_zip_no_broadcast_ListOffsetArray_dict():
    a = ak.Array([[1], []])
    b = ak.Array([[2], []])
    c = ak.zip_no_broadcast({"a": a, "b": b})
    assert ak.to_list(c) == ak.to_list(ak.zip({"a": a, "b": b}))


def test_ak_zip_no_broadcast_NumpyArray_list():
    a = ak.Array([1])
    b = ak.Array([2])
    c = ak.zip_no_broadcast([a, b])
    assert ak.to_list(c) == ak.to_list(ak.zip([a, b]))


def test_ak_zip_no_broadcast_ListOffsetArray_list():
    a = ak.Array([[1], []])
    b = ak.Array([[2], []])
    c = ak.zip_no_broadcast([a, b])
    assert ak.to_list(c) == ak.to_list(ak.zip([a, b]))


def test_typetracer_NumpyArray_non_touching():
    tracer = ak.Array([1], backend="typetracer")

    tracer, report = ak.typetracer.typetracer_with_report(
        tracer.layout.form_with_key(), highlevel=True
    )

    _ = ak.zip_no_broadcast({"foo": tracer, "bar": tracer})
    assert len(report.shape_touched) == 1
    assert len(report.data_touched) == 0


def test_typetracer_ListOffsetArray_non_touching():
    tracer = ak.Array([[1], [], [2, 3]], backend="typetracer")

    tracer, report = ak.typetracer.typetracer_with_report(
        tracer.layout.form_with_key(), highlevel=True
    )

    _ = ak.zip_no_broadcast({"foo": tracer, "bar": tracer})
    assert len(report.shape_touched) == 1
    assert len(report.data_touched) == 0



# AUTO GENERATED ON 2026-02-08 AT 18:23:02
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import cupy.testing as cpt
import numpy as np
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_1():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_2():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 3, 3]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_3():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 1, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_4():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 0, 2]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_5():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_6():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_7():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 3, 3]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_8():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 1, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_9():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 0, 2]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_10():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_11():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_12():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 3, 3]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_13():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 1, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_14():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 0, 2]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_15():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    fromoffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_16():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    fromoffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_17():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    fromoffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 3, 3]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_18():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    fromoffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 1, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_19():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    fromoffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 0, 2]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_20():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    fromoffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_21():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    fromoffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_22():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    fromoffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 3, 3]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_23():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    fromoffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 1, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_24():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    fromoffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 0, 2]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListOffsetArray_drop_none_indexes_32_25():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int32)
    noneindexes = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    fromoffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    length_offsets = 3
    length_indexes = 3
    funcC = cupy_backend['awkward_ListOffsetArray_drop_none_indexes', cupy.int32, cupy.int32, cupy.int32]
    funcC(tooffsets, noneindexes, fromoffsets, length_offsets, length_indexes)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward._meta.meta import Meta
from awkward._typing import JSONSerializable


class EmptyMeta(Meta):
    is_unknown = True
    is_leaf = True

    def purelist_parameters(self, *keys: str) -> JSONSerializable:
        return None

    @property
    def purelist_isregular(self) -> bool:
        return True

    @property
    def purelist_depth(self) -> int:
        return 1

    @property
    def is_identity_like(self) -> bool:
        return True

    @property
    def minmax_depth(self) -> tuple[int, int]:
        return (1, 1)

    @property
    def branch_depth(self) -> tuple[bool, int]:
        return (False, 1)

    @property
    def fields(self) -> list[str]:
        return []

    @property
    def is_tuple(self) -> bool:
        return False

    @property
    def dimension_optiontype(self) -> bool:
        return False



from __future__ import annotations

import numpy as np
import pytest

import awkward as ak

pd = pytest.importorskip("pandas")


_data = {
    "x": ["abc", "FG_12345"],
    "y": [None, ["g1", "g2"]],
}

_expected_inner = pd.DataFrame(
    {
        "x": ["FG_12345", "FG_12345"],
        "y": ["g1", "g2"],
    },
    index=pd.MultiIndex.from_tuples(
        [(1, 0), (1, 1)],
        names=["entry", "subentry"],
    ),
)

_expected_outer = pd.DataFrame(
    {
        "x": ["abc", "FG_12345", "FG_12345"],
        "y": [np.nan, "g1", "g2"],
    },
    index=pd.MultiIndex.from_tuples(
        [(0, 0), (1, 0), (1, 1)],
        names=["entry", "subentry"],
    ),
)

_expected_left = pd.DataFrame(
    {
        "x": ["abc", "FG_12345", "FG_12345"],
        "y": [np.nan, "g1", "g2"],
    },
    index=pd.MultiIndex.from_tuples(
        [(0, 0), (1, 0), (1, 1)],
        names=["entry", "subentry"],
    ),
)

_expected_right = pd.DataFrame(
    {
        "x": ["FG_12345", "FG_12345"],
        "y": ["g1", "g2"],
    },
    index=pd.MultiIndex.from_tuples(
        [(1, 0), (1, 1)],
        names=["entry", "subentry"],
    ),
)

params = [
    ("inner", _expected_inner),
    ("outer", _expected_outer),
    ("left", _expected_left),
    ("right", _expected_right),
]


@pytest.mark.parametrize("how, expected", params)
def test_merge_option(how: str, expected: pd.DataFrame) -> None:
    a = ak.Array(_data)
    actual = ak.to_dataframe(a, how=how)
    pd.testing.assert_frame_equal(actual, expected)



from __future__ import annotations

import numpy as np

import awkward as ak


def test_sort_record():
    a = ak.contents.NumpyArray(np.array([50, 500, 100, 1, 200, 1000]))

    record = ak.Array(ak.contents.RecordArray([a, a[::-1]], ["a", "b"]))

    assert ak.to_list(ak.sort(record)["a"]) == [1, 50, 100, 200, 500, 1000]
    assert ak.to_list(ak.sort(record)["b"]) == [1, 50, 100, 200, 500, 1000]


def test_sort_record_tuple():
    a = ak.contents.NumpyArray(np.array([50, 500, 100, 1, 200, 1000]))

    record = ak.Array(ak.contents.RecordArray([a, a[::-1]], None))

    assert ak.to_list(ak.sort(record)["0"]) == [1, 50, 100, 200, 500, 1000]
    assert ak.to_list(ak.sort(record)["1"]) == [1, 50, 100, 200, 500, 1000]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import math

import pytest

import awkward as ak


@pytest.mark.parametrize(
    "keepdims, expected_result",
    [
        pytest.param(False, ak.Array([2.25, 6.5])),
        pytest.param(True, ak.Array([[2.25], [6.5]])),
    ],
)
def test_keepdims(keepdims: bool, expected_result: ak.Array):
    data = ak.Array(
        [
            [1, 2, 3],
            [4, 7],
        ]
    )
    weight = ak.Array(
        [
            [1, 1, 2],
            [1, 5],
        ]
    )
    assert ak.all(
        ak.mean(data, weight=weight, axis=1, keepdims=keepdims) == expected_result
    )


@pytest.mark.parametrize(
    "mask_identity, expected_result",
    [
        pytest.param(False, ak.Array([1.5, math.nan, 8])),
        pytest.param(True, ak.Array([1.5, None, 8])),
    ],
)
def test_mask_identity(mask_identity: bool, expected_result: ak.Array):
    data = ak.Array(
        [
            [1, 2],
            [],
            [6, 9],
        ]
    )
    weight = ak.Array(
        [
            [1, 1],
            [],
            [1, 2],
        ]
    )
    result = ak.mean(data, weight=weight, axis=1, mask_identity=mask_identity)
    assert result[0] == expected_result[0]
    if mask_identity:
        assert result[1] is None
    else:
        assert math.isnan(result[1])  # NaN is not equal to itself per IEEE!
    assert result[2] == expected_result[2]



from __future__ import annotations

import awkward as ak


def test():
    x = ak.Array([1, 2, 3], behavior={"foo": "BAR"}, attrs={"hello": "world"})
    y = ak.Array([8, 0, 9], behavior={"do": "re"}, attrs={"mi": "fa"})
    z = x + y
    assert z.attrs == {"hello": "world", "mi": "fa"}
    assert z.behavior == {"foo": "BAR", "do": "re"}


def test_unary():
    x = ak.Array([1, 2, 3], behavior={"foo": "BAR"}, attrs={"hello": "world"})
    y = -x
    assert y.attrs == x.attrs
    assert x.behavior is y.behavior


def test_two_return():
    x = ak.Array([1, 2, 3], behavior={"foo": "BAR"}, attrs={"hello": "world"})
    y, y_ret = divmod(x, 2)
    assert y.attrs == y_ret.attrs
    assert y.attrs == x.attrs

    assert y.behavior is y_ret.behavior
    assert y.behavior is x.behavior



from __future__ import annotations

import cupy as cp
import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def test_block_boundary_sum():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(3000, size=3000)
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.sum(cuda_content, -1, highlevel=False) == ak.sum(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.sum(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.sum(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_any():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(3000, size=3000)
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.any(cuda_content, -1, highlevel=False) == ak.any(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.any(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.any(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_all():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(3000, size=3000)
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.all(cuda_content, -1, highlevel=False) == ak.all(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.all(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.all(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_sum_bool():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(2, size=3000, dtype=np.bool_)
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.sum(cuda_content, -1, highlevel=False) == ak.sum(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.sum(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.sum(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_max():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(3000, size=3000)
    print(array)
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    print(ak.max(content, -1, highlevel=False))
    print(ak.max(cuda_content, -1, highlevel=False))
    assert ak.max(cuda_content, -1, highlevel=False) == ak.max(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.max(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.max(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_min():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(3000, size=3000)
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.min(cuda_content, -1, highlevel=False) == ak.min(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.min(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.min(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_negative_min():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(3000, size=3000) * -1
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.min(cuda_content, -1, highlevel=False) == ak.min(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.min(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.min(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_argmin():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(3000, size=3000)
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.argmin(cuda_content, -1, highlevel=False) == ak.argmin(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.argmin(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.argmin(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_argmax():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(3000, size=3000)
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.argmax(cuda_content, -1, highlevel=False) == ak.argmax(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.argmax(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.argmax(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_count():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(3000, size=3000)
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.count(cuda_content, -1, highlevel=False) == ak.count(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.count(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.count(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_count_nonzero():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(2, size=3000)
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.count_nonzero(cuda_content, -1, highlevel=False) == ak.count_nonzero(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.count_nonzero(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.count_nonzero(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod():
    primes = [x for x in range(2, 30000) if all(x % n != 0 for n in range(2, x))]
    content = ak.contents.NumpyArray(primes)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.prod(cuda_content, -1, highlevel=False) == ak.prod(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.prod(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.prod(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_bool():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(2, size=3000, dtype=np.bool_)
    content = ak.contents.NumpyArray(array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.prod(cuda_content, -1, highlevel=False) == ak.prod(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.prod(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.prod(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1



from __future__ import annotations

import cupy as cp
import cupy.testing as cpt
import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def prod(xs):
    out = 1
    for x in xs:
        out *= x
    return out


def test_0115_generic_reducer_operation_sumprod_types():
    array = np.array([[True, False, False], [True, False, False]])
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)

    depth1 = ak.to_backend(depth1, "cuda")

    assert sum(to_list(np.sum(array, axis=-1))) == sum(
        to_list(ak.sum(depth1, axis=-1, highlevel=False))
    )
    assert prod(to_list(np.prod(array, axis=-1))) == prod(
        to_list(ak.prod(depth1, axis=-1, highlevel=False))
    )
    del depth1


def test_0115_generic_reducer_operation_sumprod_types_1():
    array = np.array([[0, 1, 2], [3, 4, 5]], dtype=np.int8)
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)

    depth1 = ak.to_backend(depth1, "cuda")

    assert (
        np.sum(array, axis=-1).dtype
        == ak.to_numpy(ak.sum(depth1, axis=-1, highlevel=False)).dtype
    )
    assert (
        np.prod(array, axis=-1).dtype
        == ak.to_numpy(ak.prod(depth1, axis=-1, highlevel=False)).dtype
    )
    assert sum(to_list(np.sum(array, axis=-1))) == sum(
        to_list(ak.sum(depth1, axis=-1, highlevel=False))
    )
    assert prod(to_list(np.prod(array, axis=-1))) == prod(
        to_list(ak.prod(depth1, axis=-1, highlevel=False))
    )
    del depth1


def test_0115_generic_reducer_operation_sumprod_types_2():
    array = np.array([[0, 1, 2], [3, 4, 5]], dtype=np.uint8)
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)

    depth1 = ak.to_backend(depth1, "cuda")

    assert (
        np.sum(array, axis=-1).dtype
        == ak.to_numpy(ak.sum(depth1, axis=-1, highlevel=False)).dtype
    )
    assert (
        np.prod(array, axis=-1).dtype
        == ak.to_numpy(ak.prod(depth1, axis=-1, highlevel=False)).dtype
    )
    assert sum(to_list(np.sum(array, axis=-1))) == sum(
        to_list(ak.sum(depth1, axis=-1, highlevel=False))
    )
    assert prod(to_list(np.prod(array, axis=-1))) == prod(
        to_list(ak.prod(depth1, axis=-1, highlevel=False))
    )
    del depth1


def test_0115_generic_reducer_operation_sumprod_types_3():
    array = np.array([[0, 1, 2], [3, 4, 5]], dtype=np.int16)
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)

    depth1 = ak.to_backend(depth1, "cuda")

    assert (
        np.sum(array, axis=-1).dtype
        == ak.to_numpy(ak.sum(depth1, axis=-1, highlevel=False)).dtype
    )
    assert (
        np.prod(array, axis=-1).dtype
        == ak.to_numpy(ak.prod(depth1, axis=-1, highlevel=False)).dtype
    )
    assert sum(to_list(np.sum(array, axis=-1))) == sum(
        to_list(ak.sum(depth1, axis=-1, highlevel=False))
    )
    assert prod(to_list(np.prod(array, axis=-1))) == prod(
        to_list(ak.prod(depth1, axis=-1, highlevel=False))
    )
    del depth1


def test_0115_generic_reducer_operation_sumprod_types_4():
    array = np.array([[0, 1, 2], [3, 4, 5]], dtype=np.uint16)
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)

    depth1 = ak.to_backend(depth1, "cuda")

    assert (
        np.sum(array, axis=-1).dtype
        == ak.to_numpy(ak.sum(depth1, axis=-1, highlevel=False)).dtype
    )
    assert (
        np.prod(array, axis=-1).dtype
        == ak.to_numpy(ak.prod(depth1, axis=-1, highlevel=False)).dtype
    )
    assert sum(to_list(np.sum(array, axis=-1))) == sum(
        to_list(ak.sum(depth1, axis=-1, highlevel=False))
    )
    assert prod(to_list(np.prod(array, axis=-1))) == prod(
        to_list(ak.prod(depth1, axis=-1, highlevel=False))
    )


def test_0115_generic_reducer_operation_sumprod_types_5():
    array = np.array([[0, 1, 2], [3, 4, 5]], dtype=np.int32)
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)

    depth1 = ak.to_backend(depth1, "cuda")

    assert (
        np.sum(array, axis=-1).dtype
        == ak.to_numpy(ak.sum(depth1, axis=-1, highlevel=False)).dtype
    )
    assert (
        np.prod(array, axis=-1).dtype
        == ak.to_numpy(ak.prod(depth1, axis=-1, highlevel=False)).dtype
    )
    assert sum(to_list(np.sum(array, axis=-1))) == sum(
        to_list(ak.sum(depth1, axis=-1, highlevel=False))
    )
    assert prod(to_list(np.prod(array, axis=-1))) == prod(
        to_list(ak.prod(depth1, axis=-1, highlevel=False))
    )
    del depth1


def test_0115_generic_reducer_operation_sumprod_types_6():
    array = np.array([[0, 1, 2], [3, 4, 5]], dtype=np.uint32)
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)

    depth1 = ak.to_backend(depth1, "cuda")

    assert (
        np.sum(array, axis=-1).dtype
        == ak.to_numpy(ak.sum(depth1, axis=-1, highlevel=False)).dtype
    )
    assert (
        np.prod(array, axis=-1).dtype
        == ak.to_numpy(ak.prod(depth1, axis=-1, highlevel=False)).dtype
    )
    assert sum(to_list(np.sum(array, axis=-1))) == sum(
        to_list(ak.sum(depth1, axis=-1, highlevel=False))
    )
    assert prod(to_list(np.prod(array, axis=-1))) == prod(
        to_list(ak.prod(depth1, axis=-1, highlevel=False))
    )
    del depth1


def test_0115_generic_reducer_operation_sumprod_types_7():
    array = np.array([[0, 1, 2], [3, 4, 5]], dtype=np.int64)
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)

    depth1 = ak.to_backend(depth1, "cuda")

    assert (
        np.sum(array, axis=-1).dtype
        == ak.to_numpy(ak.sum(depth1, axis=-1, highlevel=False)).dtype
    )
    assert (
        np.prod(array, axis=-1).dtype
        == ak.to_numpy(ak.prod(depth1, axis=-1, highlevel=False)).dtype
    )
    assert sum(to_list(np.sum(array, axis=-1))) == sum(
        to_list(ak.sum(depth1, axis=-1, highlevel=False))
    )
    assert prod(to_list(np.prod(array, axis=-1))) == prod(
        to_list(ak.prod(depth1, axis=-1, highlevel=False))
    )
    del depth1


def test_0115_generic_reducer_operation_sumprod_types_8():
    array = np.array([[0, 1, 2], [3, 4, 5]], dtype=np.uint64)
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)

    depth1 = ak.to_backend(depth1, "cuda")

    assert (
        np.sum(array, axis=-1).dtype
        == ak.to_numpy(ak.sum(depth1, axis=-1, highlevel=False)).dtype
    )
    assert (
        np.prod(array, axis=-1).dtype
        == ak.to_numpy(ak.prod(depth1, axis=-1, highlevel=False)).dtype
    )
    assert sum(to_list(np.sum(array, axis=-1))) == sum(
        to_list(ak.sum(depth1, axis=-1, highlevel=False))
    )
    assert prod(to_list(np.prod(array, axis=-1))) == prod(
        to_list(ak.prod(depth1, axis=-1, highlevel=False))
    )
    del depth1


def test_0115_generic_reducer_operation_sumprod_types_FIXME():
    array = np.array([[True, False, False], [True, False, False]])
    content2 = ak.contents.NumpyArray(array.reshape(-1))
    offsets3 = ak.index.Index64(np.array([0, 3, 3, 5, 6], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda")

    assert (
        np.sum(array, axis=-1).dtype
        == ak.to_numpy(ak.sum(depth1, axis=-1, highlevel=False)).dtype
    )
    assert (
        np.prod(array, axis=-1).dtype
        == ak.to_numpy(ak.prod(depth1, axis=-1, highlevel=False)).dtype
    )
    del depth1


def test_2020_reduce_axis_none_sum():
    array = ak.Array(
        [[0, 2, 3.0], [4, 5, 6, 7, 8], [], [9, 8, None], [10, 1], []], backend="cuda"
    )
    cpt.assert_allclose(ak.sum(array, axis=None), 63.0)
    assert ak.almost_equal(
        ak.sum(array, axis=None, keepdims=True),
        ak.to_regular(ak.Array([[63.0]], backend="cuda")),
    )

    arr = ak.Array([[63.0]], backend="cuda")
    assert ak.almost_equal(
        ak.sum(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(arr.mask[ak.Array([[True]], backend="cuda")]),
    )
    assert ak.sum(array[2], axis=None, mask_identity=True) is None
    del array


def test_2020_reduce_axis_none_prod():
    array = ak.Array(
        [[0, 2, 3.0], [4, 5, 6, 7, 8], [], [9, 8, None], [10, 1], []], backend="cuda"
    )
    cpt.assert_allclose(ak.prod(array[1:], axis=None), 4838400.0)
    assert ak.prod(array, axis=None) == 0
    assert ak.almost_equal(
        ak.prod(array, axis=None, keepdims=True),
        ak.to_regular(ak.Array([[0.0]], backend="cuda")),
    )
    assert ak.almost_equal(
        ak.prod(array[1:], axis=None, keepdims=True),
        ak.to_regular(ak.Array([[4838400.0]], backend="cuda")),
    )

    arr = ak.Array([[4838400.0]], backend="cuda")
    assert ak.almost_equal(
        ak.prod(array[1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(arr.mask[ak.Array([[True]], backend="cuda")]),
    )
    assert ak.prod(array[2], axis=None, mask_identity=True) is None
    del array


def test_2020_reduce_axis_none_min():
    array = ak.Array(
        [[0, 2, 3.0], [4, 5, 6, 7, 8], [], [9, 8, None], [10, 1], []], backend="cuda"
    )
    cpt.assert_allclose(ak.min(array, axis=None), 0.0)
    assert ak.almost_equal(
        ak.min(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular(ak.Array([[0.0]], backend="cuda")),
    )
    assert ak.almost_equal(
        ak.min(array, axis=None, keepdims=True, initial=-100.0, mask_identity=False),
        ak.to_regular(ak.Array([[-100.0]], backend="cuda")),
    )

    arr = ak.Array([[0.0]], backend="cuda")
    assert ak.almost_equal(
        ak.min(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(arr.mask[ak.Array([[True]], backend="cuda")]),
    )

    arr = ak.Array(ak.Array([[np.inf]], backend="cuda"))
    assert ak.almost_equal(
        ak.min(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(arr.mask[ak.Array([[False]], backend="cuda")]),
    )
    assert ak.min(array[2], axis=None, mask_identity=True) is None
    del array


def test_2020_reduce_axis_none_max():
    array = ak.Array(
        [[0, 2, 3.0], [4, 5, 6, 7, 8], [], [9, 8, None], [10, 1], []], backend="cuda"
    )
    cpt.assert_allclose(ak.max(array, axis=None), 10.0)
    assert ak.almost_equal(
        ak.max(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular(ak.Array([[10.0]], backend="cuda")),
    )
    assert ak.almost_equal(
        ak.max(array, axis=None, keepdims=True, initial=100.0, mask_identity=False),
        ak.to_regular(ak.Array([[100.0]], backend="cuda")),
    )

    arr = ak.Array([[10.0]], backend="cuda")
    assert ak.almost_equal(
        ak.max(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(arr.mask[ak.Array([[True]], backend="cuda")]),
    )

    arr = ak.Array(ak.Array([[np.inf]], backend="cuda"))
    assert ak.almost_equal(
        ak.max(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(arr.mask[ak.Array([[False]], backend="cuda")]),
    )
    assert ak.max(array[2], axis=None, mask_identity=True) is None
    del array


def test_2020_reduce_axis_none_count():
    array = ak.Array(
        [[0, 2, 3.0], [4, 5, 6, 7, 8], [], [9, 8, None], [10, 1], []], backend="cuda"
    )
    assert ak.count(array, axis=None) == 12
    assert ak.almost_equal(
        ak.count(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular(ak.Array([[12]], backend="cuda")),
    )

    arr = ak.Array([[12]], backend="cuda")
    assert ak.almost_equal(
        ak.count(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(arr.mask[ak.Array([[True]], backend="cuda")]),
    )

    arr = ak.Array([[0]], backend="cuda")
    assert ak.almost_equal(
        ak.count(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(arr.mask[ak.Array([[False]], backend="cuda")]),
    )
    assert ak.count(array[2], axis=None, mask_identity=True) is None
    assert ak.count(array[2], axis=None, mask_identity=False) == 0
    del array


def test_2020_reduce_axis_none_count_nonzero():
    array = ak.Array(
        [[0, 2, 3.0], [4, 5, 6, 7, 8], [], [9, 8, None], [10, 1], []], backend="cuda"
    )
    assert ak.count_nonzero(array, axis=None) == 11
    assert ak.almost_equal(
        ak.count_nonzero(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular(ak.Array([[11]], backend="cuda")),
    )

    arr = ak.Array([[11]], backend="cuda")
    assert ak.almost_equal(
        ak.count_nonzero(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(arr.mask[ak.Array([[True]], backend="cuda")]),
    )

    arr = ak.Array([[0]], backend="cuda")
    assert ak.almost_equal(
        ak.count_nonzero(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(arr.mask[ak.Array([[False]], backend="cuda")]),
    )
    assert ak.count_nonzero(array[2], axis=None, mask_identity=True) is None
    assert ak.count_nonzero(array[2], axis=None, mask_identity=False) == 0
    del array


def test_2020_reduce_axis_none_std_no_mask_axis_none():
    array = ak.Array(
        [[0, 2, 3.0], [4, 5, 6, 7, 8], [], [9, 8, None], [10, 1], []], backend="cuda"
    )
    out1 = ak.std(array[-1:], axis=None, keepdims=True, mask_identity=True)

    arr = ak.Array([[0.0]], backend="cuda")
    out2 = ak.to_regular(arr.mask[ak.Array([[False]], backend="cuda")])
    assert ak.almost_equal(out1, out2)

    out3 = ak.std(array[2], axis=None, mask_identity=True)
    assert out3 is None
    del array
    del out1, out2, out3


def test_2020_reduce_axis_none_std():
    array = ak.Array(
        [[0, 2, 3.0], [4, 5, 6, 7, 8], [], [9, 8, None], [10, 1], []], backend="cuda"
    )
    cpt.assert_allclose(ak.std(array, axis=None), 3.139134700306227)
    cpt.assert_allclose(
        ak.std(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[3.139134700306227]]),
    )

    arr = ak.Array([[3.139134700306227]], backend="cuda")
    cpt.assert_allclose(
        ak.std(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(arr.mask[ak.Array([[True]], backend="cuda")]),
    )
    assert np.isnan(ak.std(array[2], axis=None, mask_identity=False))
    del array
